import contextlib
import os
import shutil
from pathlib import Path


def path_to(relative: str):
    return Path(__file__).parent / relative


def make_dirs(_dir: Path):
    os.makedirs(_dir, 0o777, exist_ok=True)


def delete_dir(_dir: Path):
    shutil.rmtree(_dir, ignore_errors=True)


def delete_file(file: Path):
    with contextlib.suppress(FileNotFoundError):
        os.remove(file)


def lower_first_character(text: str) -> str:
    return text[0].lower() + text[1:]



import logging
import sys
from pathlib import Path

import click
import prompt_toolkit as pt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import FileHistory
from rich.console import Console
from rich.status import Status

from qwikcrud import __version__
from qwikcrud.generator import FastAPIAppGenerator
from qwikcrud.logger import setup_logging
from qwikcrud.provider.openai import OpenAIProvider


@click.command()
@click.option(
    "-o", "--output-dir", default=".", help="Output directory for the generated app."
)
def main(output_dir: str) -> None:
    setup_logging()

    history = Path().home() / ".qwikcrud-prompt-history.txt"
    session = pt.PromptSession(history=FileHistory(str(history)))

    ai = OpenAIProvider()
    code_generator = FastAPIAppGenerator(Path(output_dir).resolve())

    console = Console()
    console.print(
        f"[bold green]Qwikcrud v{__version__}[/bold green] - {ai.get_name()}\n\n"
        "Type [blink]/exit[/blink] at any time to exit the generator.\n"
    )

    is_first_prompt = True
    while True:
        prompt = session.prompt(
            f"qwikcrud ({'Describe your app' if is_first_prompt else 'Specify any modifications or enhancements'}) ➤ ",
            auto_suggest=AutoSuggestFromHistory(),
        )
        if prompt == "/exit":
            return
        try:
            with Status(
                f"[dim]Asking {ai.get_name()} …[/dim]", console=console
            ) as status:
                app = ai.query(prompt)
                status.update("[dim]Generating the app[/dim]")
                code_generator.clean()
                code_generator.generate(app)
            console.print(f"App successfully generated in {Path(output_dir).resolve()}")
            console.print("\nHere is the summary of the generated app:\n")
            app.summary()
            console.print(
                "Follow the instructions in the README.md file to run your application."
            )
            console.print(
                "\nYou can request changes if the generated app doesn't meet your expectations."
            )
            is_first_prompt = False
        except Exception as e:
            logging.exception(e)
            console.print("[red]Something went wrong, Try again![/red]")


if __name__ == "__main__":
    sys.exit(main())



import logging
from typing import Any, Dict, List

from openai import OpenAI

from qwikcrud.helpers import path_to
from qwikcrud.provider.base import AIProvider
from qwikcrud.schemas import App
from qwikcrud.settings import settings


class OpenAIProvider(AIProvider):
    def __init__(self):
        super().__init__()
        self.client = OpenAI(api_key=settings.openai_api_key)
        with open(path_to("prompts/system")) as f:
            system_message = f.read()
        self.messages: List[Dict[str, Any]] = [
            {
                "role": "system",
                "content": system_message,
            },
        ]

    def get_name(self) -> str:
        return f"ChatGPT ({settings.openai_model})"

    def query(self, prompt: str) -> App:
        self.messages.append({"role": "user", "content": prompt})
        completion = self.client.chat.completions.create(
            model=settings.openai_model,
            response_format={"type": "json_object"},
            messages=self.messages,
            temperature=0.4,
        )
        self.messages.append(
            {
                "role": "assistant",
                "content": completion.choices[0].message.content,
            }
        )
        logging.debug(f"Result from OpenAI: {completion.choices[0].message.content}")
        return App.model_validate_json(
            completion.choices[0].message.content, strict=False
        )



import logging

from qwikcrud.settings import settings


def setup_logging():
    logging.basicConfig(
        format="%(levelname)s [%(asctime)s] %(name)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=settings.logging_level,
    )



from abc import abstractmethod
from pathlib import Path

import autoflake
import black
import isort
from jinja2 import Environment, FileSystemLoader

from qwikcrud import helpers as h
from qwikcrud.schemas import App


class BaseAppGenerator:
    def __init__(self, output_directory: Path) -> None:
        self.env = Environment(  # noqa: S701
            loader=FileSystemLoader(h.path_to("templates")),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self.output_directory = output_directory

    def _write_code_into_file(self, path, code_text: str, format_code: bool = True):
        with open(self.output_directory / f"{path}", "w") as file:
            if format_code:
                code_text = black.format_str(code_text, mode=black.Mode())
                code_text = autoflake.fix_code(
                    code_text, remove_all_unused_imports=True
                )
                code_text = isort.code(code_text)
            file.write(code_text)

    def _generate_from_template(
        self,
        app: App,
        relative_template_path,
        destination_path=None,
        template_data=None,
        skip_render=False,
        extension="py",
    ):
        if template_data is None:
            template_data = {"app": app}
        if destination_path is None:
            destination_path = relative_template_path
        if skip_render:
            template_text = open(
                f"{h.path_to('templates')}/{self._absolute_template_path(f'{relative_template_path}.{extension}')}"
            ).read()
            self._write_code_into_file(
                f"{destination_path}.{extension}", template_text, extension == "py"
            )
        else:
            template = self.env.get_template(
                self._absolute_template_path(f"{relative_template_path}.{extension}.j2")
            )
            rendered_code = template.render(template_data)
            self._write_code_into_file(
                f"{destination_path}.{extension}", rendered_code, extension == "py"
            )

    @abstractmethod
    def _absolute_template_path(self, relative_path: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def generate(self, app: App) -> None:
        raise NotImplementedError

    @abstractmethod
    def clean(self):
        raise NotImplementedError


class FastAPIAppGenerator(BaseAppGenerator):
    def _absolute_template_path(self, relative_path: str) -> str:
        return f"fastapi/{relative_path}"

    def generate(self, app: App) -> None:
        h.make_dirs(self.output_directory / "app")
        h.make_dirs(self.output_directory / "templates")
        h.make_dirs(self.output_directory / "static/css")
        self._generate_from_template(app, "app/__init__")
        self._generate_from_template(app, "app/models")
        self._generate_from_template(app, "app/schemas")
        self._generate_from_template(app, "app/crud")
        self._generate_from_template(app, "app/deps")
        self._generate_from_template(app, "app/settings")
        self._generate_from_template(app, "app/db")
        self._generate_from_template(app, "app/pre_start")
        self._generate_from_template(app, "app/main")
        self._generate_from_template(app, "app/admin")
        self._generate_from_template(app, "requirements", extension="txt")
        self._generate_from_template(app, "README", extension="md")
        self._generate_from_template(
            app, "templates/index", extension="html", skip_render=True
        )
        self._generate_from_template(
            app, "static/css/style", extension="css", skip_render=True
        )
        self.__generate_endpoints(app)
        if app.has_file():
            self._generate_from_template(app, "app/storage")
        if app.has_enum():
            self._generate_from_template(app, "app/enums")
        self._write_code_into_file(
            ".qwikcrud.json.lock", app.model_dump_json(indent=4), format_code=False
        )

    def __generate_endpoints(self, app: App) -> None:
        h.make_dirs(self.output_directory / "app/endpoints")
        self._generate_from_template(app, "app/endpoints/__init__")
        for entity in app.entities:
            template_path = "app/endpoints/template"
            destination_path = f"app/endpoints/{entity.name.lower()}"
            self._generate_from_template(
                app, template_path, destination_path, {"entity": entity, "app": app}
            )

    def clean(self):
        h.delete_dir(self.output_directory / "app")
        h.delete_dir(self.output_directory / "templates")
        h.delete_dir(self.output_directory / "static")
        h.delete_file(self.output_directory / "requirements.txt")
        h.delete_file(self.output_directory / "README.md")



import time

from qwikcrud.helpers import path_to
from qwikcrud.provider.base import AIProvider
from qwikcrud.schemas import App


class DummyAIProvider(AIProvider):
    def get_name(self) -> str:
        return "DummyAI"

    def query(self, prompt: str) -> App:  # noqa ARG002
        time.sleep(0.1)
        return App.model_validate_json(open(path_to("../examples/app.json")).read())



from abc import abstractmethod

from qwikcrud.schemas import App


class AIProvider:
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def query(self, prompt: str) -> App:
        raise NotImplementedError



from typing import Optional

from pydantic import Field
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    openai_api_key: Optional[str] = Field(None)
    openai_model: str = "gpt-3.5-turbo-1106"
    logging_level: str = "ERROR"


settings = Settings()



from enum import Enum
from typing import Any, List, Optional

from pydantic import BaseModel, Field, model_validator
from rich.console import Console
from rich.tree import Tree

from qwikcrud import helpers as h


class FieldType(str, Enum):
    ID = "ID"
    Integer = "Integer"
    Float = "Float"
    Boolean = "Boolean"
    Date = "Date"
    Time = "Time"
    DateTime = "DateTime"
    String = "String"
    Text = "Text"
    Enum = "Enum"
    Email = "Email"
    JSON = "JSON"
    Image = "Image"
    File = "File"

    @classmethod
    def _missing_(cls: "FieldType", value: str) -> Optional["str"]:
        """When an invalid or case-mismatched string value is provided, this method attempts to find a
        case-insensitive match among the enum members."""
        value = value.lower()
        for member in cls:
            if member.lower() == value:
                return member
        return None


field_python_type_mapping = {
    FieldType.ID: "int",
    FieldType.Integer: "int",
    FieldType.Float: "float",
    FieldType.Boolean: "bool",
    FieldType.Date: "datetime.date",
    FieldType.Time: "datetime.time",
    FieldType.DateTime: "datetime.datetime",
    FieldType.String: "str",
    FieldType.Text: "str",
    FieldType.Enum: "str",
    FieldType.Email: "EmailStr",
    FieldType.JSON: "dict",
    FieldType.Image: "Union[File, UploadFile, None]",
    FieldType.File: "Union[File, UploadFile, None]",
}


class RelationType(str, Enum):
    ONE_TO_ONE = "ONE_TO_ONE"
    ONE_TO_MANY = "ONE_TO_MANY"
    MANY_TO_MANY = "MANY_TO_MANY"

    @classmethod
    def _missing_(cls: "FieldType", value: str) -> Optional["str"]:
        """When an invalid or case-mismatched string value is provided, this method attempts to find a
        case-insensitive match among the enum members."""
        value = value.lower()
        for member in cls:
            if member.lower() == value:
                return member
        return None


class Constraints(BaseModel):
    unique: Optional[bool] = Field(None)
    not_null: Optional[bool] = Field(None)
    gt: Optional[float] = Field(None)
    ge: Optional[float] = Field(None)
    lt: Optional[float] = Field(None)
    le: Optional[float] = Field(None)
    multiple_of: Optional[float] = Field(None)
    min_length: Optional[int] = Field(None)
    max_length: Optional[int] = Field(None)
    mime_types: Optional[List[str]] = Field(None)
    allowed_values: Optional[List[str]] = Field(None)

    def get_allowed_values(self):
        return self.allowed_values or []


class FieldModel(BaseModel):
    name: str = Field(...)
    type_: FieldType = Field(..., alias="type")
    constraints: Optional[Constraints] = Field(Constraints())

    @model_validator(mode="after")
    def root_validator(self) -> "FieldModel":
        self.name = self.name.lower()
        return self

    def is_id(self):
        return self.type_ == FieldType.ID

    def is_file(self):
        return self.type_ in (FieldType.Image, FieldType.File)

    def sqla_column_def(self) -> str:
        """Generate the sqlalchemy column definition

        Example:
            id: Mapped[int] = mapped_column(primary_key=True)
        """
        type_mapping = field_python_type_mapping.get(self.type_)
        if self.constraints.not_null:
            type_mapping = f"Optional[{type_mapping}]"
        mapped_column_kwargs = []
        if self.type_ in [FieldType.Image, FieldType.File]:
            file_field_mapping = (
                "ImageField" if self.type_ == FieldType.Image else "FileField"
            )
            file_field_mapping_kwargs = []
            if self.type_ == FieldType.Image:
                file_field_mapping_kwargs.append("thumbnail_size=(150,150)")
            file_field_validators = ['SizeValidator(max_size="20M")']
            if self.constraints.mime_types:
                mimetypes_list = ",".join(
                    ('"' + v + '"') for v in self.constraints.mime_types
                )
                file_field_validators.append(
                    f"ContentTypeValidator([{mimetypes_list}])"
                )
            file_field_mapping_kwargs.append(
                f'validators=[{",".join(file_field_validators)}]'
            )
            file_field_mapping += f'({",".join(file_field_mapping_kwargs)})'
            mapped_column_kwargs.append(file_field_mapping)
        if self.type_ == FieldType.Enum:
            mapped_column_kwargs.append(f"Enum({self.name.capitalize()})")
        if self.constraints.unique:
            mapped_column_kwargs.append("unique=True")
        if self.type_ == FieldType.ID:
            mapped_column_kwargs.append("primary_key=True")
        return f"{self.name}: Mapped[{type_mapping}]" + (
            f'=mapped_column({",".join(mapped_column_kwargs)})'
            if len(mapped_column_kwargs) > 0
            else ""
        )

    def pydantic_def(self, all_optional: bool = False):
        type_mapping = field_python_type_mapping.get(self.type_)
        if self.type_ == FieldType.Enum:
            type_mapping = self.name.capitalize()
        if all_optional or self.constraints.not_null:
            type_mapping = f"Optional[{type_mapping}]"
        if self.is_file():
            type_mapping = "Optional[FileInfo]"
        pydantic_field_kwargs = [
            f"{k}={v}"
            for (k, v) in self.constraints.model_dump(
                exclude={"unique", "not_null", "allowed_values"}, exclude_none=True
            ).items()
        ]
        if all_optional:
            pydantic_field_kwargs = ["None", *pydantic_field_kwargs]

        return f"{self.name}: {type_mapping}" + (
            f'=Field({",".join(pydantic_field_kwargs)})'
            if len(pydantic_field_kwargs) > 0
            else ""
        )


class Entity(BaseModel):
    name: str = Field(...)
    fields: List[FieldModel] = Field(...)

    @model_validator(mode="after")
    def root_validator(self) -> "Entity":
        self.name = self.name.capitalize()
        return self


class Relation(BaseModel):
    name: str
    type_: RelationType = Field(..., alias="type")
    from_: str = Field(..., alias="from")
    to: str
    field_name: str
    backref_field_name: str

    @model_validator(mode="after")
    def root_validator_mode_after(self) -> "Relation":
        self.name = self.name.lower()
        self.from_ = self.from_.capitalize()
        self.to = self.to.capitalize()
        self.field_name = h.lower_first_character(self.field_name)
        self.backref_field_name = h.lower_first_character(self.backref_field_name)
        return self

    @model_validator(mode="before")
    @classmethod
    def root_validator_mode_before(cls, data: Any) -> Any:
        if isinstance(data, dict):
            # Transform MANY_TO_ONE to ONE_TO_MANY
            if data.get("type", "ONE_TO_ONE").upper() == "MANY_TO_ONE":
                data["type"] = "ONE_TO_MANY"
                data["from"], data["to"] = data["to"], data["from"]
                data["field_name"], data["backref_field_name"] = (
                    data["backref_field_name"],
                    data["field_name"],
                )
        return data


class App(BaseModel):
    name: str = Field(...)
    description: str = Field(...)
    entities: List[Entity] = Field(...)
    relations: List[Relation] = Field(...)

    def has_file(self) -> bool:
        for entity in self.entities:
            for field in entity.fields:
                if field.is_file():
                    return True
        return False

    def has_enum(self) -> bool:
        for entity in self.entities:
            for field in entity.fields:
                if field.type_ == FieldType.Enum:
                    return True
        return False

    def summary(self):
        console = Console()
        console.print(f"[bold magenta]Name[/bold magenta]: {self.name}")
        console.print(f"[bold magenta]Description[/bold magenta]: {self.description}")
        console.print("\n[bold underline]Entities:[/bold underline]\n")

        for entity in self.entities:
            entity_tree = Tree(f"[bold cyan]{entity.name}[/bold cyan]")
            for field in entity.fields:
                field_str = f"[cyan]{field.name}[/cyan]: [yellow]{field.type_}[/yellow]"
                constraints_str = ", ".join(
                    f"{key}={value}"
                    for key, value in field.constraints.model_dump().items()
                    if value is not None
                )
                if constraints_str:
                    field_str += f" ({constraints_str})"
                entity_tree.add(field_str)

            console.print(entity_tree)
            console.print("\n")

        if self.relations:
            console.print("[bold underline]Relationships:[/bold underline]\n")

            for relation in self.relations:
                console.print(
                    f"[bold cyan]{relation.from_}[/bold cyan] ([magenta]{relation.field_name}[/magenta])"
                    f" --[{relation.type_}]--> [bold cyan]{relation.to}[/bold cyan]"
                    f" ([magenta]{relation.backref_field_name}[/magenta])"
                )
            console.print("\n")
